/*++

Copyright (c) 1996  Microsoft Corporation

Module Name:

    adapter.c

Abstract:

    routines for binding/unbinding to/from underlying miniport drivers

Author:

    Jim Mateer 4-1-97

Environment:

    Kernel Mode

Revision History:

--*/

#include "ImSamp.h"
#pragma hdrstop

PADAPTER CurrentAdapter;


VOID
BindToLowerMP(
	OUT	PNDIS_STATUS			Status,
	IN	NDIS_HANDLE				BindContext,
	IN	PNDIS_STRING			MPDeviceName,
	IN	PVOID					SystemSpecific1,
	IN	PVOID					SystemSpecific2
	);

STATIC NDIS_STATUS
GetAdapterRegistryData(
    PNDIS_STRING IMParamsKey,
    PADAPTER Adapter
    );


STATIC VOID
ProcessLowerMPOpenAdapter(
    IN  PADAPTER Adapter,
    IN  NDIS_STATUS Status
    );

VOID
LowerMPOpenAdapterComplete(
    IN  NDIS_HANDLE ProtocolBindingContext,
    IN  NDIS_STATUS Status,
    IN  NDIS_STATUS OpenErrorStatus
    );

STATIC NDIS_STATUS
AllocatePacketPool(
    PADAPTER Adapter                   
    );

STATIC NDIS_STATUS
AllocateReceiveBufferPools(
    PADAPTER Adapter
    );



PADAPTER
FindAdapterByName(
    PWCHAR AdapterName
    );

NDIS_STATUS
MPInitialize(
	OUT PNDIS_STATUS			OpenErrorStatus,
	OUT PUINT					SelectedMediumIndex,
	IN	PNDIS_MEDIUM			MediumArray,
	IN	UINT					MediumArraySize,
	IN	NDIS_HANDLE				MiniportAdapterHandle,
	IN	NDIS_HANDLE				WrapperConfigurationContext
	);

VOID
UnbindFromLowerMP(
	OUT	PNDIS_STATUS			Status,
	IN	NDIS_HANDLE				ProtocolBindingContext,
	IN	NDIS_HANDLE				UnbindContext
	);

VOID
DerefAdapter(
    PADAPTER Adapter
    );

VOID
CleanupAdapter(
    PADAPTER Adapter
    );

VOID
LowerMPCloseAdapterComplete(
    IN  NDIS_HANDLE ProtocolBindingContext,
    IN  NDIS_STATUS Status
    );

VOID
CLUnloadProtocol(
    VOID
    );

VOID
MPHalt(
	IN	NDIS_HANDLE				MiniportAdapterContext
	);

NDIS_STATUS
MPReset(
	OUT PBOOLEAN				AddressingReset,
	IN	NDIS_HANDLE				MiniportAdapterContext
	);

#if DBG
VOID
DumpAdapter(
    PADAPTER Adapter
    );
#endif


#pragma NDIS_PAGEABLE_FUNCTION(BindToLowerMP)
#pragma NDIS_PAGEABLE_FUNCTION(GetAdapterRegistryData)

VOID
BindToLowerMP(
	OUT	PNDIS_STATUS			Status,
	IN	NDIS_HANDLE				BindContext,
	IN	PNDIS_STRING			MPDeviceName,
	IN	PVOID					SystemSpecific1,
	IN	PVOID					SystemSpecific2
	)

/*++

Routine Description:

    Bind to the underlying MP. Allocate space for an adapter structure,
    initializing its fields. Try to open the adapter indicated in MPDeviceName.

Arguments:

    See the DDK...

Return Values:

    None

--*/

{
    PADAPTER Adapter;
	NDIS_STATUS OpenAdapterStatus;
	NDIS_STATUS OpenErrorStatus;
    NDIS_STATUS LocalStatus;
    NDIS_MEDIUM MediumArray[] = {
        NdisMediumFddi,
        NdisMedium802_5,
        NdisMedium802_3,
		NdisMediumWan
        };
	UINT MediumArraySize = sizeof( MediumArray ) / sizeof( NDIS_MEDIUM );
    UINT MediaIndex;
    ULONG AdapterStructSize;
    

    PAGED_CODE();

#if BINARY_COMPATIBLE
    ImDbgOut( DBG_INFO, DBG_PROTOCOL | DBG_INIT, ("BindToLowerMP: %s\n", MPDeviceName->Buffer ));
#else
    ImDbgOut( DBG_INFO, DBG_PROTOCOL | DBG_INIT, ("BindToLowerMP: %ws\n", MPDeviceName->Buffer ));
#endif

    //
    // allocate enough space for the structure and two unicode buffers to hold
    // the "IM_" and underlying MP device names. We add 3 extra Unicode char to
    // to the IM device name to hold the  IM_ addition to the MP name,  that is be appended later
    // on and another unicode char to separate the two strings for reading .
	// The IM adapter will have the form '\Device\IM_IEEPRO3' for example, if it is sitting on top of 
	// IEEPRO3
    //

    AdapterStructSize = sizeof( ADAPTER ) + MPDeviceName->Length      // space for the miniport device name
                                          + MPDeviceName->Length +	  // space for the IM virtual adapter name ( IM_<MpDeviceName>)
										  4 * sizeof( UNICODE_NULL );

    IMAllocatePool( Adapter, AdapterStructSize, Adapter );

    if ( Adapter == NULL ) {
        PWCHAR StringData[2];

        StringData[0] = IMDriverName.Buffer;
        StringData[1] = L"Adapter";
        NdisWriteErrorLogEntry(IMDriverObject,
                               (ULONG)EVENT_TRANSPORT_RESOURCE_POOL,
                               0,
                               2,
                               &StringData,
                               0,
                               NULL);

        *Status = NDIS_STATUS_RESOURCES;
        return;
    }

    //
    // set our state to initializing
    //

    Adapter->IMMPState = ADAPTER_STATE_INITIALIZING;

    //
    // make the device name point to the buffer allocated at the back of the
    // structure. The MPDevice name looks like \Device\EE163. This is the lower adapter that we want
	// to bind to. We want to construct an intermediate device that looks like \Device\IM_EE163 and create an adapter
	// instance by this name
    //

    Adapter->IMDeviceName.MaximumLength = MPDeviceName->MaximumLength + 3 * sizeof( UNICODE_NULL );// 3 for the IM_
    Adapter->IMDeviceName.Length = Adapter->IMDeviceName.MaximumLength;
    Adapter->IMDeviceName.Buffer = (PWSTR)( Adapter + 1 );

    Adapter->MPDeviceName.MaximumLength = MPDeviceName->Length;
    Adapter->MPDeviceName.Length = Adapter->MPDeviceName.MaximumLength;
    Adapter->MPDeviceName.Buffer = (PWSTR)((PCHAR)Adapter->IMDeviceName.Buffer +
                                           Adapter->IMDeviceName.MaximumLength +
                                           sizeof( UNICODE_NULL ));


	 RtlCopyMemory(Adapter->IMDeviceName.Buffer,  // first set up the \Device\IM_  portion	
	               L"\\Device\\IM_", 
	               sizeof(L"\\Device\\IM_")
	              ); 

	 // now tack on the devices name. It is passed in as \Device\EE163. Since we 
	// already have the \\Device\ part, all we need is the EE163 part

	 RtlCopyMemory(&(Adapter->IMDeviceName.Buffer[sizeof("\\Device\\IM")]),
	 &(MPDeviceName->Buffer[sizeof("\\Device")]),
	 MPDeviceName->Length - sizeof(L"\\Device")
	 );
	  


   // LocalStatus = GetAdapterRegistryData( (PNDIS_STRING)SystemSpecific1, Adapter );

    if (0) {

#if BINARY_COMPATIBLE
        ImDbgOut(DBG_FAILURE, DBG_PROTOCOL | DBG_INIT,
                 ("(%08X) BindToLowerMP: Couldn't get registry data %08X (%s)\n",
                  Adapter, LocalStatus, MPDeviceName->Buffer ));
#else
        ImDbgOut(DBG_FAILURE, DBG_PROTOCOL | DBG_INIT,
                 ("(%08X) BindToLowerMP: Couldn't get registry data %08X (%ws)\n",
                  Adapter, LocalStatus, MPDeviceName->Buffer ));
#endif

        *Status = NDIS_STATUS_FAILURE;

        IMFreePool( Adapter );
        return;

    }

    //
    // init the event now since we use it in the completion handler
    //
    // remember our binding context so we can complete BindAdapter later on
    //

    NdisInitializeEvent( &Adapter->BlockingEvent );
    Adapter->BindContext = BindContext;


	// Open the adapter below us.
    NdisOpenAdapter(&OpenAdapterStatus,
                    &OpenErrorStatus,
                    &Adapter->LowerMPHandle,
					&MediaIndex,
                    MediumArray,
					MediumArraySize,
                    ClientProtocolHandle,
                    Adapter,
                    MPDeviceName,
                    0,
                    NULL);

    //
    // Memphis doesn't like you to do NdisRequests from within the OpenAdapterComplete
    // handler so wait for OpenAdapterComplete to finish and then do the follow up
    // processing.
    //

    if ( OpenAdapterStatus == NDIS_STATUS_PENDING ) {

        NdisWaitEvent( &Adapter->BlockingEvent, 0 );
        NdisResetEvent( &Adapter->BlockingEvent );
    } else {

        Adapter->FinalStatus = OpenAdapterStatus;
    }

    if ( NT_SUCCESS( Adapter->FinalStatus )) {

        Adapter->MediaType = MediumArray[ MediaIndex ];								 

    	//NdisWan will select the NdisMediumWan type, but RasArp needs 802.3 media type. 
		// Do the check and conversion here.
    	if(Adapter->MediaType = NdisMediumWan) Adapter->MediaType = NdisMedium802_3;  
    }

    ProcessLowerMPOpenAdapter( Adapter, Adapter->FinalStatus );

    *Status = Adapter->FinalStatus;

} // BindToLowerMP


STATIC NDIS_STATUS
GetAdapterRegistryData(
    PNDIS_STRING IMParamsKey,
    PADAPTER Adapter
    )

/*++

Routine Description:

    Obtain the IMSamp specific info associated with the underlying MP

Arguments:

    IMParamsKey - location in the registry

    Adapter - pointer to adapter context struct

Return Value:

    NDIS_STATUS_SUCCESS if everything worked ok

--*/

{
    NDIS_STATUS Status;
    NDIS_HANDLE ConfigHandle;
    NDIS_STRING IMInstanceNumberKey = NDIS_STRING_CONST( "InstanceNumber" );
    PNDIS_CONFIGURATION_PARAMETER ConfigParam;
    ULONG i;
    PWCHAR StringData[2];
    PVOID DumpData;

    PAGED_CODE();

    NdisOpenProtocolConfiguration( &Status, &ConfigHandle, IMParamsKey );

    if ( !NT_SUCCESS( Status )) {

#if BINARY_COMPATIBLE
        ImDbgOut(DBG_FAILURE, DBG_PROTOCOL | DBG_INIT,
                 ( "(%08X) GetAdapterRegistryData: can't open key %s (%08X)\n",
                   Adapter, IMParamsKey->Buffer, Status ));
#else
        ImDbgOut(DBG_FAILURE, DBG_PROTOCOL | DBG_INIT,
                 ( "(%08X) GetAdapterRegistryData: can't open key %ws (%08X)\n",
                   Adapter, IMParamsKey->Buffer, Status ));
#endif

        return Status;
    }

    //
    // get the IM device instance number and build the device instance string
    //

    NdisReadConfiguration(&Status,
                          &ConfigParam,
                          ConfigHandle,
                          &IMInstanceNumberKey,
                          NdisParameterInteger);


    if ( !NT_SUCCESS( Status )) {

        ImDbgOut(DBG_FAILURE, DBG_PROTOCOL | DBG_INIT,
                 ("(%08X) GetAdapterRegistryData: Missing InstanceNumber key\n", Adapter));

        Status = NDIS_STATUS_FAILURE;
        goto CloseConfig;
    }

    Adapter->DevInstance = (USHORT)ConfigParam->ParameterData.IntegerData;

    NdisMoveMemory( Adapter->IMDeviceName.Buffer, IMMPName.Buffer, IMMPName.Length );

    Adapter->IMDeviceName.Buffer[ IMMPName.Length / sizeof( WCHAR ) ] = L'0' + Adapter->DevInstance;


CloseConfig:
    NdisCloseConfiguration( ConfigHandle );

    return Status;

} // GetAdapterRegistryData


STATIC VOID
ProcessLowerMPOpenAdapter(
    IN  PADAPTER Adapter,
    IN  NDIS_STATUS Status
    )

/*++

Routine Description:

    Complete the binding on the lower miniport. Initialize the adapter structure, query
    the MP for certain funtionality and initialize the associated IM miniport device

Arguments:

    see the DDK

Return Values:

    None

--*/

{
    NTSTATUS EventStatus;
    PWSTR SecondaryName;
    NDIS_HARDWARE_STATUS HWStatus;
    NDIS_MEDIA_STATE MediaState = 0xFFFFFFFF;
    NDIS_STRING IMDevName;
    ULONG MacOptions;
    ULONG ErrorLogData[2];
    PWCHAR StringData[2];
    PVOID DumpData;

    ImDbgOut(DBG_TRACE, DBG_PROTOCOL, ("(%08X) ProcessLowerMPOpenAdapter\n", Adapter));

    //
    // see if it worked; clean up and get out
    //

    if ( !NT_SUCCESS( Status )) {

        ImDbgOut(DBG_FAILURE,
                 DBG_PROTOCOL,
                 ("(%08X) ProcessLowerMPOpenAdapter: binding failed %08X\n", Adapter, Status));

        if ( Status == NDIS_STATUS_ADAPTER_NOT_FOUND ) {

            EventStatus = EVENT_TRANSPORT_ADAPTER_NOT_FOUND;

        } else {

            EventStatus = EVENT_TRANSPORT_BINDING_FAILED;
        }

        StringData[0] = Adapter->IMDeviceName.Buffer;
        StringData[1] = Adapter->MPDeviceName.Buffer;
        DumpData = &Status;

        NdisWriteErrorLogEntry(IMDriverObject,
                               EventStatus,
                               0,
                               2,
                               &StringData,
                               sizeof( Status ),
                               DumpData);

        IMFreePool( Adapter );
        return;
    }

    //
    // initialize the variables associated with this structure. From this point on,
    // DerefAdapter must be called if any errors are encountered.
    //

    NdisAllocateSpinLock( &Adapter->Lock );
    Adapter->RefCount = 1;
    ImDbgOut( DBG_TRACE, DBG_REFCNTS, ("(%08X) =1 Adapter\n", Adapter ));
    InitializeListHead( &Adapter->ClientList );
    Adapter->ShutdownMask = 0;


    //
    // add adapter on list of known adapters
    //

    NdisInterlockedInsertTailList( &AdapterList, &Adapter->Linkage, &AdapterListLock );

    //
    // determine if adapter is ready for use. All adapters probably support hardware
    // status. If they don't, then we'll assume it must be ready for use. If the
    // HW status indicates ok, then check for media connect state which is new for 4.0.
    // If the media is not connected, then we wait until we get an indication that it
    // is connected and then we query the adapter for its link speed. This is important
    // for multi-speed adapters, like 10/100 ethernet.
    //
     //

    Status = MakeLocalNdisRequest(Adapter,
                                  OID_GEN_HARDWARE_STATUS,
                                  &HWStatus,
                                  sizeof( HWStatus ),
                                  NULL);

    if ( Status == NDIS_STATUS_INVALID_OID || HWStatus == NdisHardwareStatusReady ) {

        Status = MakeLocalNdisRequest(Adapter,
                                      OID_GEN_MEDIA_CONNECT_STATUS,
                                      &MediaState,
                                      sizeof( MediaState ),
                                      NULL);

        if ( Status == NDIS_STATUS_INVALID_OID || MediaState == NdisMediaStateConnected ) {

            Status = MakeLocalNdisRequest(Adapter,
                                          OID_GEN_LINK_SPEED,
                                          &Adapter->LinkSpeed,
                                          sizeof( Adapter->LinkSpeed ),
                                          NULL);

            if ( !NT_SUCCESS( Status )) {

                ImDbgOut(DBG_INFO, DBG_PROTOCOL,
                         ("(%08X) ProcessLowerMPOpenAdapter: Can't get link speed - Status %08X\n",
                          Adapter, Status));
        
                ErrorLogData[ 0 ] = IM_ERROR_MISSING_OID;
                ErrorLogData[ 1 ] = OID_GEN_LINK_SPEED;

                NdisWriteErrorLogEntry(Adapter->LowerMPHandle,
                                       NDIS_ERROR_CODE_MISSING_CONFIGURATION_PARAMETER,
                                       2,
                                       ErrorLogData);

                goto ErrorExit;
            }

         } else {

            ImDbgOut(DBG_INFO, DBG_PROTOCOL,
                     ("(%08X) ProcessLowerMPOpenAdapter: Media not connected\n", Adapter ));
        }
    } else {

        ImDbgOut(DBG_INFO, DBG_PROTOCOL,
                 ("(%08X) ProcessLowerMPOpenAdapter: HW Status not ready (%d)\n", HWStatus));
    }

    //
    // determine whether we can copy the lookahead data directly
    //

    Status = MakeLocalNdisRequest(Adapter,
                                  OID_GEN_MAC_OPTIONS,
                                  &MacOptions,
                                  sizeof( MacOptions ),
                                  NULL);

    if ( NT_SUCCESS( Status )) {

        Adapter->CopyLookaheadData = (BOOLEAN)(MacOptions & NDIS_MAC_OPTION_COPY_LOOKAHEAD_DATA);
    }

    //
    // allocate a packet pool with no buffers. We allocate all the packets and
    // stuff them on a lookaside list so we don't take the hit later on
    //

    Status = AllocatePacketPool( Adapter );

    if ( !NT_SUCCESS( Status )) {

        goto ErrorExit;
    }

    //
    // now allocate two buffer pools: one to hold header plus lookahead data and the
    // other for residual lookahead data that is copied via NdisTransferData.
    //

    Status = AllocateReceiveBufferPools( Adapter );

    if ( !NT_SUCCESS( Status )) {

        goto ErrorExit;
    }


    //
    // initialize the corresponding device instance associated with the underlying
    // miniport. NdisIMInitializeDeviceInstanceEx doesn't want the \Device part of
    // the name so we strip that off
    //

    NdisInitUnicodeString( &IMDevName, &Adapter->IMDeviceName.Buffer[ 8 ] );


    //
    // temporarily stuff the adapter pointer into a global so we can find it when
    // MiniportInitialize is called.
    CurrentAdapter = Adapter;

    Status = NdisIMInitializeDeviceInstance( LMDriverHandle, &IMDevName );

    if ( !NT_SUCCESS( Status )) {
        
#if BINARY_COMPATIBLE
        ImDbgOut(DBG_FAILURE, DBG_PROTOCOL,
                 ("(%08X) ProcessLowerMPOpenAdapter: can't init IM device %s (%08X)\n",
                  Adapter, IMDevName.Buffer, Status));
#else
        ImDbgOut(DBG_FAILURE, DBG_PROTOCOL,
                 ("(%08X) ProcessLowerMPOpenAdapter: can't init IM device %ws (%08X)\n",
                  Adapter, IMDevName.Buffer, Status));
#endif

        ErrorLogData[ 0 ] = IM_ERROR_CANT_INITIALIZE_IMSAMP_DEVICE;
        ErrorLogData[ 1 ] = Status;

        NdisWriteErrorLogEntry(Adapter->LowerMPHandle,
                               NDIS_ERROR_CODE_DRIVER_FAILURE,
                               2,
                               ErrorLogData);

        goto ErrorExit;
    }

    Adapter->ShutdownMask |= SHUTDOWN_DEINIT_DEV_INSTANCE;

    return;

    //
    // when things go wrong...
    //

ErrorExit:

    DerefAdapter( Adapter );

} // ProcessLowerMPOpenAdapter


VOID
LowerMPOpenAdapterComplete(
    IN  PADAPTER Adapter,
    IN  NDIS_STATUS Status,
    IN  NDIS_STATUS OpenErrorStatus
    )

/*++

Routine Description:

    Signal that the binding on the lower miniport is complete

Arguments:

    see the DDK

Return Values:

    None

--*/

{
    NTSTATUS EventStatus;
    PWSTR SecondaryName;
    NDIS_HARDWARE_STATUS HWStatus;
    NDIS_MEDIA_STATE MediaState = 0xFFFFFFFF;
    NDIS_STRING PSDevName;
    ULONG MacOptions;
    ULONG ErrorLogData[2];
    PWCHAR StringData[2];
    PVOID DumpData;

    ImDbgOut(DBG_TRACE, DBG_PROTOCOL, ("(%08X) LowerMPOpenAdapterComplete\n", Adapter));

    //
    // stuff the final status in the Adapter block and signal the bind handler to continue
    //

    Adapter->FinalStatus = Status;
    NdisSetEvent( &Adapter->BlockingEvent );

} // LowerMPOpenAdapterComplete


STATIC NDIS_STATUS
AllocatePacketPool(
    PADAPTER Adapter                   
    )

/*++

Routine Description:

    Allocate the packet pool. Note there are no buffers associated with this pool.

Arguments:

    Adapter - pointer to adapter context block

Return Value:

    None

--*/

{
    NDIS_STATUS Status;
    PNDIS_PACKET Packet;
    ULONG ErrorLogData[2];
    ULONG i;
    ULONG ProtoReservedSize;

    NdisInitializeSListHead( &Adapter->PacketSList );
    NdisAllocateSpinLock( &Adapter->PacketSListLock );

    //
    // figure the amount of space we need in the protocol reserved area.
    //

    ProtoReservedSize =  sizeof( IM_PACKET_CONTEXT );

    NdisAllocatePacketPool(&Status,
                           &Adapter->PacketPoolHandle,
                           ConfigData.PacketPoolSize,
                           ProtoReservedSize);

    if ( NT_SUCCESS( Status )) {

        Adapter->ShutdownMask |= SHUTDOWN_DEALLOC_PACKET_POOL;

        for ( i = 0; i < ConfigData.PacketPoolSize; ++i ) {

            NdisAllocatePacket( &Status, &Packet, Adapter->PacketPoolHandle );

            if ( !NT_SUCCESS( Status )) {

                ErrorLogData[ 0 ] = IM_ERROR_PACKET;
                ErrorLogData[ 1 ] = Status;

                NdisWriteErrorLogEntry(Adapter->LowerMPHandle,
                                       NDIS_ERROR_CODE_OUT_OF_RESOURCES,
                                       2,
                                       ErrorLogData);

                return Status;
            }

            IMSetTag( IM_PACKET_CONTEXT_FROM_PACKET( Adapter, Packet ), PacketContext );
            NdisInterlockedPushEntrySList(&Adapter->PacketSList,
                                          (PSINGLE_LIST_ENTRY)Packet->ProtocolReserved,
                                          &Adapter->PacketSListLock);
        }
    } else {

        ErrorLogData[ 0 ] = IM_ERROR_PACKET_POOL;
        ErrorLogData[ 1 ] = Status;

        NdisWriteErrorLogEntry(Adapter->LowerMPHandle,
                               NDIS_ERROR_CODE_OUT_OF_RESOURCES,
                               2,
                               ErrorLogData);
    }

    return Status;

}   // AllocatePacketPool


STATIC NDIS_STATUS
AllocateReceiveBufferPools(
    PADAPTER Adapter
    )

/*++

Routine Description:

    Allocate two buffer pools, one used for header and lookahead data and the other
    for residual lookahead data

Arguments:

    Adapter - pointer to adapter context block

Return Value:

    None

--*/

{
    NDIS_STATUS Status;
    ULONG i;
    PVOID SysBuffer;
    PNDIS_BUFFER NdisBuffer;
    PIM_BUFFER_CONTEXT BufContext;
    ULONG HeaderSize;
    ULONG FrameSize;            // doesn't include the header
    NDIS_ERROR_CODE ErrorCode;
    ULONG ErrorLogData[2];

    //
    // max amount of data w/o the MAC header
    //

    Status = MakeLocalNdisRequest(Adapter,
                                  OID_GEN_MAXIMUM_FRAME_SIZE,
                                  &FrameSize,
                                  sizeof( FrameSize ),
                                  NULL);

    if ( !NT_SUCCESS( Status )) {

        ImDbgOut(DBG_FAILURE, DBG_PROTOCOL,
                 ("(%08X) AllocateReceiveBufferPool: Can't get frame size - Status %08X\n",
                  Adapter, Status));
        
        ErrorCode = NDIS_ERROR_CODE_MISSING_CONFIGURATION_PARAMETER;
        ErrorLogData[ 0 ] = IM_ERROR_MISSING_OID;
        ErrorLogData[ 1 ] = OID_GEN_MAXIMUM_FRAME_SIZE;

        goto ErrorExit;
    }

    //
    // this one includes the header
    //

    Status = MakeLocalNdisRequest(Adapter,
                                  OID_GEN_MAXIMUM_TOTAL_SIZE,
                                  &Adapter->TotalSize,
                                  sizeof( Adapter->TotalSize ),
                                  NULL);

    if ( !NT_SUCCESS( Status )) {

        ImDbgOut(DBG_FAILURE, DBG_PROTOCOL,
                 ("(%08X) AllocateReceiveBufferPool: Can't get total size - Status %08X\n",
                  Adapter, Status));
        
        ErrorCode = NDIS_ERROR_CODE_MISSING_CONFIGURATION_PARAMETER;
        ErrorLogData[ 0 ] = IM_ERROR_MISSING_OID;
        ErrorLogData[ 1 ] = OID_GEN_MAXIMUM_TOTAL_SIZE;

        goto ErrorExit;
    }

    //
    // figure the real header size
    //

    HeaderSize = Adapter->TotalSize - FrameSize;

    Status = MakeLocalNdisRequest(Adapter,
                                  OID_GEN_MAXIMUM_LOOKAHEAD,
                                  &Adapter->LookaheadBufferSize,
                                  sizeof( Adapter->LookaheadBufferSize ),
                                  NULL);

    if ( !NT_SUCCESS( Status )) {

        ImDbgOut(DBG_FAILURE, DBG_PROTOCOL,
                 ("(%08X) AllocateReceiveBufferPool: Can't get lookahead size - Status %08X\n",
                  Adapter, Status));
        
        ErrorCode = NDIS_ERROR_CODE_MISSING_CONFIGURATION_PARAMETER;
        ErrorLogData[ 0 ] = IM_ERROR_MISSING_OID;
        ErrorLogData[ 1 ] = OID_GEN_MAXIMUM_LOOKAHEAD;

        goto ErrorExit;
    }

    //
    // figure the buffer size for the lookahead buffers, which also include
    // the header. The residual buffer size is the total size, since the MP
    // can indicate less than the max lookahead buffer size, we need to be
    // prepared to receive the whole frame in the residual buffer. On the
    // other hand, the MP can indicate more than the lookahead size, so we
    // might bypass the lookahead buffer and just use a residual buffer on
    // the receive.
    //

    Adapter->LookaheadBufferSize += HeaderSize;

    //
    // init the lookahead buffer pool
    //

    NdisInitializeSListHead( &Adapter->LookaheadSList );
    NdisAllocateSpinLock( &Adapter->LookaheadSListLock );

    NdisAllocateBufferPool(&Status,
                           &Adapter->LookaheadPoolHandle,
                           ConfigData.LookaheadPoolSize);

    if ( !NT_SUCCESS( Status )) {

        ErrorCode = NDIS_ERROR_CODE_OUT_OF_RESOURCES;
        ErrorLogData[ 0 ] = IM_ERROR_LOOKAHEAD_POOL;
        ErrorLogData[ 1 ] = Status;

        goto ErrorExit;
    }

    Adapter->ShutdownMask |= SHUTDOWN_DEALLOC_LOOKAHEAD_POOL;

    for ( i = 0; i < ConfigData.LookaheadPoolSize; ++i ) {

        IMAllocatePoolNoStructTag( SysBuffer, Adapter->LookaheadBufferSize, LookaheadBuffer );

        if ( SysBuffer == NULL ) {

            ErrorCode = NDIS_ERROR_CODE_OUT_OF_RESOURCES;
            ErrorLogData[ 0 ] = IM_ERROR_VM_LOOKAHEAD_BUFFER;
            ErrorLogData[ 1 ] = (ULONG)NDIS_STATUS_RESOURCES;

            goto ErrorExit;
        }

        NdisAllocateBuffer(&Status,
                           &NdisBuffer,
                           Adapter->LookaheadPoolHandle,
                           SysBuffer,
                           Adapter->LookaheadBufferSize);

        if ( !NT_SUCCESS( Status )) {

            ErrorCode = NDIS_ERROR_CODE_OUT_OF_RESOURCES;
            ErrorLogData[ 0 ] = IM_ERROR_LOOKAHEAD_BUFFER;
            ErrorLogData[ 1 ] = Status;

            goto ErrorExit;
        }

        //
        // use the VM area to store a pointer back to the NDIS buffer and link it
        // onto the S List.
        //

        BufContext = (PIM_BUFFER_CONTEXT)SysBuffer;
        BufContext->NdisBuffer = NdisBuffer;

        NdisInterlockedPushEntrySList(&Adapter->LookaheadSList,
                                      &BufContext->SListEntry,
                                      &Adapter->LookaheadSListLock);
    }

    //
    // same thing for the residual buffer pool
    //

    NdisInitializeSListHead( &Adapter->ResidualSList );
    NdisAllocateSpinLock( &Adapter->ResidualSListLock );

    NdisAllocateBufferPool(&Status,
                           &Adapter->ResidualPoolHandle,
                           ConfigData.ResidualPoolSize);

    if ( !NT_SUCCESS( Status )) {

        ErrorCode = NDIS_ERROR_CODE_OUT_OF_RESOURCES;
        ErrorLogData[ 0 ] = IM_ERROR_RESIDUAL_POOL;
        ErrorLogData[ 1 ] = Status;

        goto ErrorExit;
    }

    Adapter->ShutdownMask |= SHUTDOWN_DEALLOC_RESIDUAL_POOL;

    for ( i = 0; i < ConfigData.ResidualPoolSize; ++i ) {

        IMAllocatePoolNoStructTag( SysBuffer, Adapter->TotalSize, ResidualBuffer );

        if ( SysBuffer == NULL ) {

            ErrorCode = NDIS_ERROR_CODE_OUT_OF_RESOURCES;
            ErrorLogData[ 0 ] = IM_ERROR_VM_RESIDUAL_BUFFER;
            ErrorLogData[ 1 ] = (ULONG)NDIS_STATUS_RESOURCES;

            goto ErrorExit;
        }

        NdisAllocateBuffer(&Status,
                           &NdisBuffer,
                           Adapter->ResidualPoolHandle,
                           SysBuffer,
                           Adapter->TotalSize);

        if ( !NT_SUCCESS( Status )) {

            ErrorCode = NDIS_ERROR_CODE_OUT_OF_RESOURCES;
            ErrorLogData[ 0 ] = IM_ERROR_RESIDUAL_BUFFER;
            ErrorLogData[ 1 ] = Status;

            goto ErrorExit;
        }

        //
        // use the VM area to store a pointer back to the NDIS buffer and link it
        // onto the S List.
        //

        BufContext = (PIM_BUFFER_CONTEXT)SysBuffer;
        BufContext->NdisBuffer = NdisBuffer;

        NdisInterlockedPushEntrySList(&Adapter->ResidualSList,
                                      &BufContext->SListEntry,
                                      &Adapter->ResidualSListLock);
    }

    return Status;

    //
    // when things go awry
    //

ErrorExit:

    NdisWriteErrorLogEntry( Adapter->LowerMPHandle, ErrorCode, 2, ErrorLogData );
    return Status;

}   // AllocateReceiveBufferPool






NDIS_STATUS
MPInitialize(
	OUT PNDIS_STATUS			OpenErrorStatus,
	OUT PUINT					SelectedMediumIndex,
	IN	PNDIS_MEDIUM			MediumArray,
	IN	UINT					MediumArraySize,
	IN	NDIS_HANDLE				MiniportAdapterHandle,
	IN	NDIS_HANDLE				WrapperConfigurationContext
	)

/*++

Routine Description:

    Imsample's device initialization routine. The list of media types is
    checked to be sure it is one that we support. If so, match up the name of
    the device being opened with one of the adapters to which we've bound.

Arguments:

    See the DDK...

Return Values:

    None

--*/

{
    NDIS_HANDLE ConfigHandle;
    NDIS_STATUS Status;
    NDIS_STRING LowerAdapterKey = NDIS_STRING_CONST( "LowerAdapter" );
    PNDIS_CONFIGURATION_PARAMETER ConfigParam;
    NDIS_STRING LowerAdapter;
    PLIST_ENTRY NextAdapter;
    PADAPTER AdapterInList;
    ULONG ErrorLogData[2];
	PNDIS_MINIPORT_BLOCK Mp = (PNDIS_MINIPORT_BLOCK)MiniportAdapterHandle;
    ImDbgOut(DBG_TRACE, DBG_MINIPORT, ("MPInitialize: Enter\n"));
    ImDbgOut( DBG_INFO, DBG_MINIPORT | DBG_INIT, ("MiniportInitialize Miniport->BaseName =  %ws\n",Mp->MiniportName.Buffer ));





    // Find the adapter we are trying to init. When we bind to a real adapter below, we create an Adapter structure, and put it on our
	// list of bound adapters. Then we create an Intermediate adapter to export to the protocols above us. When NDIS calls our
	// initialization function, we have to find the adapter in our list that we are trying to initialize. NDIS passes us a 
	// Miniport handle, that has the name of the adapter we want to init. Our adapter structure also holds the name of the adapter 
	// we are trying to creat. All we have to do is match up the two names.
	// Look, Ma, No Registry!


 	AdapterInList = FindAdapterByName(Mp->MiniportName.Buffer);	// Returns NULL if an adapter is not found. I can't think 
	                                                            // of anyway this could happen, so if it does, it is bad.


    // check that we really found a matching adapter
    //

    if ( !AdapterInList ) {

//        // This means we have an IM MP entry that does not match up with an actual adapter. This is bad.   

//BUG BUG change the error code to reflect the new IM device naming convention

        ImDbgOut(DBG_FAILURE, DBG_MINIPORT | DBG_INIT, ("Can't find adapter for MP dev # %ws\n",Mp->MiniportName.Buffer));
 
        ErrorLogData[ 0 ] = IM_ERROR_BAD_REGISTRY_DATA;
        ErrorLogData[ 1 ] = IM_ERROR_INVALID_IMSAMP_MP_INSTANCE;

        NdisWriteErrorLogEntry(MiniportAdapterHandle,
                               NDIS_ERROR_CODE_MISSING_CONFIGURATION_PARAMETER,
                               2,
                               ErrorLogData);

        return NDIS_STATUS_FAILURE;
    }

    //
    // lookup our media type in the supplied media array
    //

    for (--MediumArraySize ; MediumArraySize > 0; ) {

        if ( MediumArray[ MediumArraySize ] == AdapterInList->MediaType ) {

            break;
        }

        if ( MediumArraySize == 0 ) {

            break;
        }

        --MediumArraySize;
    }

    if ( MediumArraySize == 0 && MediumArray[ 0 ] != AdapterInList->MediaType ) {

        return NDIS_STATUS_UNSUPPORTED_MEDIA;
    }

    *SelectedMediumIndex = MediumArraySize;

    //
    // save NDIS's handle to us
    //

    AdapterInList->IMNdisHandle = MiniportAdapterHandle;

    //
    // finish the initialization process by set our attributes
    //

    NdisMSetAttributesEx(MiniportAdapterHandle,
                         AdapterInList,
                         0,
                         NDIS_ATTRIBUTE_DESERIALIZE|
                         NDIS_ATTRIBUTE_IGNORE_PACKET_TIMEOUT|
                         NDIS_ATTRIBUTE_IGNORE_REQUEST_TIMEOUT  ,
                         0);
                         
   	// mark the adapter as running
   AdapterInList->IMMPState = ADAPTER_STATE_RUNNING;

 
    return NDIS_STATUS_SUCCESS;

} // MPInitialize




PADAPTER
FindAdapterByName(
    PWCHAR AdapterName
    )

/*++

Routine Description:

    Using the name, find the associated adapter struct. Up the
    adapter's ref count before leaving

Arguments:

    AdapterName - pointer to unicode string version of adapter name

Return Value:

    pointer to ADAPTER struct, otherwise NULL

--*/

{
    PLIST_ENTRY NextAdapter;
    PADAPTER AdapterInList;
    ULONG NameLength = 0;
    PWCHAR pw = AdapterName;

    while ( *pw++ != 0  && NameLength < 64 ) {
        ++NameLength;
    }
   
    NameLength *= sizeof( WCHAR );

    NdisAcquireSpinLock( &AdapterListLock );

    NextAdapter = AdapterList.Flink;
    while ( NextAdapter != &AdapterList ) {

        AdapterInList = CONTAINING_RECORD( NextAdapter, ADAPTER, Linkage );

        if ( AdapterInList->IMDeviceName.Length == (NameLength+2) ) { // allow for the UNICODE NULL that our counting alg. did not count

            if ( NdisEqualMemory(AdapterInList->IMDeviceName.Buffer,
                                 AdapterName,
                                 NameLength)) {
                break;
            }
        }

        NextAdapter = NextAdapter->Flink;
    }

    if ( NextAdapter != &AdapterList ) {

        NdisAcquireSpinLock( &AdapterInList->Lock );

        ++AdapterInList->RefCount;
        ImDbgOut( DBG_TRACE, DBG_REFCNTS, ("(%08X) +Adapter:k %2d\n",
                                           AdapterInList, AdapterInList->RefCount ));

        NdisReleaseSpinLock( &AdapterInList->Lock );
    } else {

        AdapterInList = NULL;
    }

    NdisReleaseSpinLock( &AdapterListLock );

    return AdapterInList;
}


VOID
UnbindFromLowerMP(
	OUT	PNDIS_STATUS			Status,
	IN	NDIS_HANDLE				ProtocolBindingContext,
	IN	NDIS_HANDLE				UnbindContext
	)

/*++

Routine Description:

    Called by NDIS to indicate that an adapter is going away. Release our reference
    on the adapter and set the closing flag to true to prevent any further
    references from being obtained.

Arguments:

    See the DDK...

Return Values:

    None

--*/

{
    PADAPTER Adapter = (PADAPTER)ProtocolBindingContext;
    NDIS_STATUS LocalStatus;

    ImDbgOut(DBG_TRACE, DBG_PROTOCOL | DBG_INIT, ("(%08X) UnbindFromLowerMP\n", Adapter));

    IMStructAssert( Adapter );

    //
    // set the adapter to closing
    //

    NdisAcquireSpinLock( &Adapter->Lock );
    Adapter->IMMPState = ADAPTER_STATE_CLOSING;
    NdisReleaseSpinLock( &Adapter->Lock );

    //
    // deinit the IM miniport associated with the underlying MP
    //

    if ( Adapter->ShutdownMask & SHUTDOWN_DEINIT_DEV_INSTANCE ) {

        LocalStatus = NdisIMDeInitializeDeviceInstance( Adapter->IMNdisHandle );
        IMAssert( NT_SUCCESS( LocalStatus ));

        Adapter->ShutdownMask &= ~SHUTDOWN_DEINIT_DEV_INSTANCE;
    }

    //
    // Remember the UnbindContext so we can complete the unbinding explicitly.
    // Remove the reference on the adapter.
    //

    Adapter->BindContext = UnbindContext;

    DerefAdapter( Adapter );

    *Status = NDIS_STATUS_PENDING;

} // UnbindAdapter


VOID
DerefAdapter(
    PADAPTER Adapter
    )

/*++

Routine Description:

    Decrement the ref counter associated with this structure. When it goes to
    zero, close the adapter, and delete the memory associated with the struct

Arguments:

    Adapter - pointer to adapter context block

Return Value:

    number of references remaining associated with this structure

--*/

{
    LONG RefCount;

    IMStructAssert( Adapter );

    NdisAcquireSpinLock( &Adapter->Lock );
	
	if(Adapter->RefCount){
         RefCount = --Adapter->RefCount;
	    
    ImDbgOut( DBG_TRACE, DBG_REFCNTS, ("(%08X) -Adapter:a %2d\n", Adapter, Adapter->RefCount ));

    }
    
    if ( RefCount == 0 ) {

        CleanupAdapter( Adapter );
    } else {

        NdisReleaseSpinLock( &Adapter->Lock );
    }

    IMAssert ( RefCount >= 0 );
} // DerefAdapter


VOID
CleanupAdapter(
    PADAPTER Adapter
    )

/*++

Routine Description:

    All refs have been removed on this struct. Free its resources.

    Called with adapter lock HELD

Arguments:

    Adapter - pointer to adapter context block

Return Value:

    None

--*/

{
    NDIS_STATUS Status = NDIS_STATUS_SUCCESS;

    NdisReleaseSpinLock( &Adapter->Lock );

    //
    // set the adapter to closing. we may have gotten here during init failure
    // so we may not have previously set the state to closing
    //

    NdisAcquireSpinLock( &Adapter->Lock );
    Adapter->IMMPState = ADAPTER_STATE_CLOSING;
    NdisReleaseSpinLock( &Adapter->Lock );

      //
    // deinit the IM miniport associated with the underlying MP
    //

    if ( Adapter->ShutdownMask & SHUTDOWN_DEINIT_DEV_INSTANCE ) {

        Status = NdisIMDeInitializeDeviceInstance( Adapter->IMNdisHandle );
        IMAssert( NT_SUCCESS( Status ));
    }

    //
    // close the adapter. remainder of processing is done in the completion
    // routine
    //

    NdisCloseAdapter( &Status, Adapter->LowerMPHandle );
    IMAssert( NT_SUCCESS( Status ));

    if ( Status != NDIS_STATUS_PENDING ) {

        LowerMPCloseAdapterComplete( Adapter, Status );
    }
} // CleanupAdapter


VOID
LowerMPCloseAdapterComplete(
    IN  NDIS_HANDLE ProtocolBindingContext,
    IN  NDIS_STATUS Status
    )

/*++

Routine Description:

    Completion routine for NdisCloseAdapter. All that should be left is to free
    the pool associated with the structure

Arguments:

    See the DDK...

Return Values:

    None

--*/

{
    PADAPTER Adapter = (PADAPTER)ProtocolBindingContext;
    PSINGLE_LIST_ENTRY SListEntry;
    PNDIS_PACKET Packet;
    PNDIS_BUFFER NdisBuffer;
    PVOID SysBuffer;
    ULONG SysLength;
    PIM_BUFFER_CONTEXT BufContext;
    BOOLEAN TimerCancelled;

    ImDbgOut(DBG_TRACE, DBG_PROTOCOL, ("(%08X) LowerMPCloseAdapterComplete\n", Adapter));

    IMStructAssert( Adapter );

    IMAssert( NT_SUCCESS( Status ));

    //
    // complete the unbinding if necessary
    //

    if ( Adapter->BindContext ) {

        NdisCompleteUnbindAdapter( Adapter->BindContext, Status );
    }

    //
    // acquire the main adapter list lock and remove this entry
    //

    NdisAcquireSpinLock( &AdapterListLock );
    RemoveEntryList( &Adapter->Linkage );
    NdisReleaseSpinLock( &AdapterListLock );

    //
    // return packet pool resources
    //

    while ( SListEntry = NdisInterlockedPopEntrySList(&Adapter->PacketSList,
                                                      &Adapter->PacketSListLock)) {

        NdisFreePacket( CONTAINING_RECORD( SListEntry, NDIS_PACKET, ProtocolReserved ));
    }

    if ( Adapter->ShutdownMask & SHUTDOWN_DEALLOC_PACKET_POOL ) {

        NdisFreePacketPool( Adapter->PacketPoolHandle );
    }

    //
    // return lookahead buffer pool resources
    //

    while ( SListEntry = NdisInterlockedPopEntrySList(&Adapter->LookaheadSList,
                                                      &Adapter->LookaheadSListLock)) {

        BufContext = (PIM_BUFFER_CONTEXT)SListEntry;

        NdisAdjustBufferLength( BufContext->NdisBuffer, Adapter->LookaheadBufferSize );
        NdisFreeBuffer( BufContext->NdisBuffer );
        IMFreePool( BufContext );
    }

    if ( Adapter->ShutdownMask & SHUTDOWN_DEALLOC_LOOKAHEAD_POOL ) {

        NdisFreeBufferPool( Adapter->LookaheadPoolHandle );
    }

    //
    // same for residual buffer pool resources
    //

    while ( SListEntry = NdisInterlockedPopEntrySList(&Adapter->ResidualSList,
                                                      &Adapter->ResidualSListLock)) {

        BufContext = (PIM_BUFFER_CONTEXT)SListEntry;

        NdisAdjustBufferLength( BufContext->NdisBuffer, Adapter->TotalSize );
        NdisFreeBuffer( BufContext->NdisBuffer );
        IMFreePool( BufContext );
    }

    if ( Adapter->ShutdownMask & SHUTDOWN_DEALLOC_RESIDUAL_POOL ) {

        NdisFreeBufferPool( Adapter->ResidualPoolHandle );
    }

    //
    // free adapter lock from dispatcher DB and free mem for adapter struct
    //

    NdisFreeSpinLock( &Adapter->Lock );

    IMFreePool( Adapter );

} // LowerMPCloseAdapterComplete


VOID
CLUnloadProtocol(
    VOID
    )

/*++

Routine Description:

    Memphis checks to see if this defined for checked builds - sigh...

Arguments:

    None

Return Value:

    None

--*/

{

}


VOID
MPHalt(
	IN	NDIS_HANDLE				MiniportAdapterContext
	)

/*++

Routine Description:

    This handler is called on Memphis. It indicates that the IM MP is no more
    and we should avoid calling NdisIMDeInitializeDeviceInstance...

Arguments:

    See the DDK...

Return Values:

    None

--*/

{
    PADAPTER Adapter = (PADAPTER)MiniportAdapterContext;

    IMStructAssert( Adapter );

    ImDbgOut(DBG_TRACE, DBG_MINIPORT, ("(%08X) MPHalt\n", Adapter));

    Adapter->ShutdownMask &= ~SHUTDOWN_DEINIT_DEV_INSTANCE;
}

NDIS_STATUS
MPReset(
	OUT PBOOLEAN				AddressingReset,
	IN	NDIS_HANDLE				MiniportAdapterContext
	)

/*++

Routine Description:



Arguments:

    See the DDK...

Return Values:

    None

--*/

{
    PADAPTER Adapter = (PADAPTER)MiniportAdapterContext;

    IMStructAssert( Adapter );

    ImDbgOut(DBG_TRACE, DBG_MINIPORT, ("(%08X) MPReset\n", Adapter));

    *AddressingReset = FALSE;

#if BINARY_COMPATIBLE
#if DBG
    //
    // linker doesn't include the routine if no one is calling it. force some code to
    // be generated for it
    //

    {
        volatile INT A = 0;
        if ( A ) {
            DumpAdapter( Adapter );
        }
    }
#endif // DBG
#endif // BINARY_COMPATIBLE

    return NDIS_STATUS_SUCCESS;
}



